


#ifndef WIN32
#include <sys/times.h>
#include <unistd.h>
#endif

#include "rtree.h"

#define DEBUG 1

void overflow(node_type *, int, int, node_type *, node_type *);


/****************************/
/* Start Utility procedures */
/****************************/

void initialize(configR_type *configR)
{ 
  FILE *fp;
  int string_len;

  fp = fopen(CONFIG_FILE, "r");

  fscanf(fp, "m=%d\n", &m);
  fscanf(fp, "M=%d\n", &M);
  fscanf(fp, "dim=%d\n", &dim);
  fscanf(fp, "reinsert_p=%d\n", &reinsert_p);
  fscanf(fp, "no_histogram=%d\n",&no_histogram);

  fgets(configR->nodefile, FILENAME_MAX, fp);
  string_len = strlen(configR->nodefile);
  configR->nodefile[string_len-1] = '\0';

  fgets(configR->rootfile, FILENAME_MAX, fp);
  string_len = strlen(configR->rootfile);
  configR->rootfile[string_len-1] = '\0';

  fgets(configR->positionfile, FILENAME_MAX, fp);
  string_len = strlen(configR->positionfile);
  configR->positionfile[string_len-1] = '\0';

  fgets(configR->queryfile, FILENAME_MAX, fp);
  string_len = strlen(configR->queryfile);
  configR->queryfile[string_len-1] = '\0';


} 
/* initialize */


void initialize_argument(configR_type *configR, char **argv)
{ 
  //FILE *fp;
  //int string_len;

  m = atoi(argv[5]);
  M = atoi(argv[6]);
  dim = atoi(argv[2]);
  reinsert_p = atoi(argv[7]);
  no_histogram = atoi(argv[1]);

  strcpy(configR->positionfile, argv[3]);
  strcpy(SAVE_RTREE_FILE, argv[4]);


} 

void tree_node_allocate(node_type **node)
{

  (*node) = (node_type *)malloc(sizeof(node_type));
  (*node)->a = (float *)malloc(sizeof(float) * dim);
  (*node)->b = (float *)malloc(sizeof(float) * dim);
  (*node)->ptr = (node_type **)malloc(sizeof(node_type *) * M);
  memset((*node)->ptr, 0, sizeof(node_type *) * M);
  (*node)->classGroup = classSet_new();

}


void tree_node_deallocate(node_type *free_node)
{

  free(free_node->a);
  free(free_node->b);
  free(free_node->ptr);

  free(free_node);

}



void cal_MBR_node_node(float *new_a, float *new_b, node_type *node1, node_type *node2)
{
  int i;

  for (i=0; i<dim; i++) {

	if (node1->a[i] < node2->a[i])
		new_a[i] = node1->a[i];
	else
		new_a[i] = node2->a[i];
  }

  for (i=0; i<dim; i++) {

        if (node1->b[i] > node2->b[i])
                new_b[i] = node1->b[i];
        else
                new_b[i] = node2->b[i];
  }

  return;

}


double cal_vol(float *a, float *b)
{
  int i;
  double volume = 1.0;

  for (i=0; i<dim; i++)
	volume = volume * (double)(b[i] - a[i]);


  return(volume);

}


double cal_overlap(node_type *node1, node_type *node2)
{
  double overlap;

  int i;

  overlap = 1.0;
  for (i=0; i<dim; i++) {

	/* 6 possible cases */

	if (node2->a[i] > node1->b[i] || node1->a[i] > node2->b[i]) {

		overlap = 0.0;
		break;

	}
	else if (node2->a[i] <= node1->a[i]) {

		if (node2->b[i] <= node1->b[i]) {

			// a2, a1, b2, b1

			overlap = overlap * (node2->b[i] - node1->a[i]);

		}
		else {

			// a2, a1, b1, b2

			overlap = overlap * (node1->b[i] - node1->a[i]);

		}
	}
	else if (node1->a[i] < node2->a[i]) {

		if (node2->b[i] <= node1->b[i]) {

			// a1, a2, b2, b1

			overlap = overlap * (node2->b[i] - node2->a[i]);

		}
		else {

			// a1, a2, b1, b2

			overlap = overlap * (node1->b[i] - node2->a[i]);

		}

	}

  }


  return(overlap);

}


double cal_overlap_sum(node_type *node, int index_skip,  node_type *parent_node)
{
  double overlap;

  int i, stop;


  overlap = 0.0;
  stop = M - parent_node->vacancy;
  for (i=0; i<stop; i++) {

	if (i == index_skip) continue;

	overlap = overlap + cal_overlap(parent_node->ptr[i], node);

  }


  return(overlap);

}


double Dist2(node_type *node1, node_type *node2)
{
  double distance;
  float *point1, *point2;

  int i;

  point1 = (float *)malloc(sizeof(float) * dim);
  point2 = (float *)malloc(sizeof(float) * dim);

  for (i=0; i<dim; i++) {
	//*** Ray has type casting the following with float type
	point1[i] = (float)((node1->b[i] - node1->a[i])/2.0);
	point2[i] = (float)((node2->b[i] - node2->a[i])/2.0);
  }

  distance = 0.0;
  for (i=0; i<dim; i++) {
	distance = distance + pow((double)(point1[i] - point2[i]), 2.0);
  }

  return(pow(distance, 0.5));

}


/**************************/
/* End Utility procedures */
/**************************/


int least_overlap_enlarge(node_type *parent_node, node_type  *data_node)
{
  double new_overlap_diff, old_overlap, new_overlap; 
  double vol_at_index, new_vol, min_overlap_diff;
  int index;
  node_type *temp_node;

  int i, stop;
 

  tree_node_allocate(&temp_node);

  stop = M - parent_node->vacancy;
  for(i=0; i<stop; i++) {

        /* original overlap */
   
        old_overlap = cal_overlap_sum(parent_node->ptr[i], i, parent_node);
 
        /* overlap after enlargement */
 
        cal_MBR_node_node(temp_node->a, temp_node->b, parent_node->ptr[i], data_node);
  
        new_overlap = cal_overlap_sum(temp_node, i, parent_node);

        /* check if index is needed to updated */
        new_overlap_diff = new_overlap - old_overlap;

	if (i == 0) {

                index = i;
                min_overlap_diff = new_overlap_diff;   

	}
	else {
        	if (new_overlap_diff < min_overlap_diff) {
                	index = i;
	                min_overlap_diff = new_overlap_diff;
        	}
        	else if (new_overlap_diff == min_overlap_diff) {

			vol_at_index = cal_vol(parent_node->ptr[index]->a, parent_node->ptr[index]->b);
			new_vol = cal_vol(temp_node->a, temp_node->b);
	                if(new_vol < vol_at_index) {

        	                index=i;

	                }
		}
       	}
        
  }  /* end i */

  tree_node_deallocate(temp_node);
        
                
  return (index);
                
}



/********************************************************/
/* least_area_enlarge():                                */
/* Select the node which will cause least bounding      */
/* box enlargement when insert an entry to it           */
/********************************************************/

int least_area_enlarge(node_type *parent_node, node_type *data_node)
{ 
  double new_vol_diff, old_vol, new_vol;
  double vol_at_index, min_vol_diff;
  int index;
  float *temp_a, *temp_b;
  int i, stop;

  temp_a = (float *)malloc(sizeof(float) * dim);
  temp_b = (float *)malloc(sizeof(float) * dim);


  stop = M - parent_node->vacancy;
  for(i=0; i<stop; i++) { 

	/* original volume */

	old_vol = cal_vol(parent_node->ptr[i]->a, parent_node->ptr[i]->b);            	

	/* volume after enlargement */

	cal_MBR_node_node(temp_a, temp_b, parent_node->ptr[i], data_node);

	new_vol = cal_vol(temp_a, temp_b);

	/* check if index is needed to updated */
        new_vol_diff = new_vol - old_vol;

	if (i == 0) {

                index = i;
                min_vol_diff = new_vol_diff;
                vol_at_index = new_vol;

	}
	else {
		if (new_vol_diff < min_vol_diff) { 
			index = i;
           		min_vol_diff = new_vol_diff;
	           	vol_at_index = new_vol;
		}
		else if (new_vol_diff == min_vol_diff) { 
			if(new_vol < vol_at_index) {
              			index=i;
              			vol_at_index = new_vol;
			}
	        }
	}
      
  }  /* end i */

  free(temp_a);
  free(temp_b);
            
  return (index);

} 
/* least_enlarge */




/**********************************************************/
/* choose_leaf():                                         */
/* Select a leaf node in which to place a new index entry */
/* current_level is the level of the current_node	  */
/**********************************************************/

int choose_leaf(node_type **node_found, node_type *current_node, int 
current_level, node_type *data_node) 
{

  int child_chosen, level_found;

  /*******/
  /* CL1 */
  /*******/
  
  /**********************************************************/
  /* Initialise: 			     		    */  
  /* Set N to be the root node(already done in insert_node) */
  /**********************************************************/

  /*  It has been already done in insert()  */

  /*******/
  /* CL2 */
  /*******/

  /**************/
  /* Leaf check */
  /**************/
  
  if(current_node->attribute == ROOT && (current_node->ptr[0]==NULL || current_node->ptr[0]->attribute==LEAF)) { 

	/************************************************************
	*node_found is the root of the tree because the root is 
	the only internal node of the tree at that time.
	*************************************************************/

   	return(current_level);		// root is at level 0				

  }

  if (current_node->ptr[0]->attribute == LEAF) {
        *node_found = current_node;
        return(current_level);
  }

  /*******/
  /* CL3 */
  /*******/

  /******************/
  /* Choose subtree */
  /******************/

  if (current_node->ptr[0]->ptr[0]->attribute != LEAF) {

	child_chosen = least_area_enlarge(current_node, data_node);

  }
  else {

	child_chosen = least_overlap_enlarge(current_node, data_node);

  }

  /*******/
  /* CL4 */
  /*******/
    
  /***********************/
  /* Descend recursively */
  /***********************/

  current_node = current_node->ptr[child_chosen];
  level_found = choose_leaf(node_found, current_node, current_level+1, data_node);


  return(level_found);

} /* choose_leaf */



void adjust_MBR(node_type *node_inserted, myClassSet *classID)
{
  node_type *node = node_inserted;
  int i, flag;

  while(node->attribute != ROOT) {
	      
  	flag = FALSE;
  	for(i=0; i < dim; i++) { 
		if(node->parent->a[i] > node->a[i]) {
			node->parent->a[i] = node->a[i];
			flag = TRUE;
		}
		if(node->parent->b[i] < node->b[i]) {
			node->parent->b[i] = node->b[i];
			flag = TRUE;
		}
  	}

	//set_unionNonRedundant(node->parent->classGroup, classID);
	classSet_unionNonRedundant(node->parent->classGroup, classID);

	if (flag == FALSE) break;

	node = node->parent;

  }

  return;

} /* adjust_MBR */



void adjust_MBR_delete(node_type *node_inserted)
{
  node_type *node = node_inserted, *parent;
  int j, stop;

  while(node->attribute != ROOT) {
  
	parent = node->parent;
	stop = M - parent->vacancy;

	for (j=0; j<dim; j++) {
		parent->a[j] = parent->ptr[0]->a[j];
		parent->b[j] = parent->ptr[0]->b[j];
	}

	//set_copy(parent->classGroup, parent->ptr[0]->classGroup);
	classSet_copy(parent->classGroup, parent->ptr[0]->classGroup);
//printf("stop %d\n", stop);
	for (j=1; j<stop; j++) {
		cal_MBR_node_node(parent->a, parent->b, parent, parent->ptr[j]);

		//set_unionNonRedundant(parent->classGroup, parent->ptr[j]->classGroup);
		classSet_unionNonRedundant(parent->classGroup, parent->ptr[j]->classGroup);
	}
  
//printf("3\n");
     //   if (flag == FALSE) break;
 
        node = parent;
        
  }
         
  return;

} /* adjust_MBR_delete */     



void swap(int *sorted_index, double *value, int i, int j)
{ 
  int temp_index;
  double temp_value; 

  temp_value = value[i];  
  value[i] = value[j];
  value[j] = temp_value;
  
  temp_index = sorted_index[i];
  sorted_index[i] = sorted_index[j];
  sorted_index[j] = temp_index;
                
  return;
   
}



int rtree_partition(int *sorted_index, double *value, int start_index, int end_index)
{
  double pivot_value;  
  int i, j;
        
  pivot_value = value[start_index];
        
  i = start_index - 1;
  j = end_index + 1;
                
  while (1) {
   
        do {    
                j = j - 1;
  
        } while (value[j] > pivot_value);
 
        do {
                i = i + 1;

        } while (value[i] < pivot_value);

        if (i < j)
                swap(sorted_index, value, i, j);
        else
                return(j);
        
  }     


  //**** Ray have added the return value -1.
  return -1;

}



/* I think it is sorted in increasing order */

void rtree_quicksort(int *sorted_index, double *value, int start_index, int end_index)
{
  int pivot;

  if (start_index < end_index) {

        pivot = rtree_partition(sorted_index, value, start_index, end_index);

        rtree_quicksort(sorted_index, value, start_index, pivot);

        rtree_quicksort(sorted_index, value, pivot+1, end_index);

  }
 
  return;

}



void sort_entries(int *sorted_index, node_type **overnode, int axis_sort)
{
  int i, start, end;
  double *value;

  //**** Ray has changed here to fix the bug
  //**** from "sizeof(double) * M+1" to "sizeof(double) * (M+1)"
  value = (double *)malloc(sizeof(double) * (M+1));


  for (i=0; i<M+1; i++) {
	sorted_index[i] = i;
	value[i] = (double)overnode[i]->a[axis_sort];
  }


  rtree_quicksort(sorted_index, value, 0, M);

  i = 0;
  while (i<M) {

	if (value[i] == value[i+1]) {

		start = i;

		while (i < M && value[i] == value[i+1]) {

			value[i] = (double)overnode[sorted_index[i]]->b[axis_sort];
			i ++;

		}

		value[i] = (double)overnode[sorted_index[i]]->b[axis_sort];
		end = i;	



		if ((end - start) > 1) {
			rtree_quicksort(sorted_index, value, start, end);
		}
		else {
			if (value[end] < value[start]) {
				swap(sorted_index, value, start, end);
			}
		}
	}


	i++;

  }

  free(value);


  return;

}


int ChooseSplitAxis(node_type **overnode)
{
  int axis_chosen, *sorted_index;
  node_type *group1, *group2;
  double new_margin_value, min_margin_value;


  int i, j, k, l, stop, cut;


  //**** Ray has corrected the following bugs
  //**** from "sizeof(int) * M+1" to "sizeof(int) * (M+1)"
  sorted_index = (int *)malloc(sizeof(int) * (M+1));
  tree_node_allocate(&group1);
  tree_node_allocate(&group2);


  for (i=0; i<dim; i++) {

	sort_entries(sorted_index, overnode, i);   // sort the entries by axis i


	new_margin_value = 0.0;
	stop = M - 2*m + 1;
	for (k=0; k<stop; k++) {

        	for (l=0; l<dim; l++) {
                	group1->a[l] = overnode[sorted_index[0]]->a[l];
	                group1->b[l] = overnode[sorted_index[0]]->b[l];
        	        group2->a[l] = overnode[sorted_index[M]]->a[l];
                	group2->b[l] = overnode[sorted_index[M]]->b[l];
        	}        


		j = 0;
		cut = m + k;
		while (j < M+1) {

			if (j < cut) {

				cal_MBR_node_node(group1->a, group1->b, group1, overnode[sorted_index[j]]);

			}
			else {

				cal_MBR_node_node(group2->a, group2->b, group2, overnode[sorted_index[j]]);

			}

			j++;

		}

		for (l=0; l<dim; l++) {

			new_margin_value = new_margin_value + group1->b[l] - group1->a[l];
			new_margin_value = new_margin_value + group2->b[l] - group2->a[l];

		}
	}

	if (i == 0) {

                axis_chosen = i;
                min_margin_value = new_margin_value;
	}
	else {

		if (new_margin_value < min_margin_value) {

			axis_chosen = i;
			min_margin_value = new_margin_value;
		}

	}	
  }

  tree_node_deallocate(group1);
  tree_node_deallocate(group2);
  free(sorted_index);

  return(axis_chosen);

} /* ChooseSplitAxis */



void ChooseSplitIndex(node_type **overnode, int axis_chosen, node_type 
*group1_chosen, node_type *group2_chosen)
{
  int split_index, *sorted_index;
  node_type *group1, *group2;  
  double new_overlap_value, min_overlap_value;
  double vol_at_index, new_vol;

  int i, j, k, stop, cut;
  
//FILE *fp;

  //**** Ray has fixed the bugs
  //**** from "sizeof(int) * M+1" to "sizeof(int) * (M+1)"
  sorted_index = (int *)malloc(sizeof(int) * (M+1));

  tree_node_allocate(&group1);
  tree_node_allocate(&group2);

  
  sort_entries(sorted_index, overnode, axis_chosen);   // sort the entries by the axis, axis_chosen


  new_overlap_value = 0.0;  
  stop = M - 2*m + 1;
  for (k=0; k<stop; k++) {

	for (i=0; i<dim; i++) {
		group1->a[i] = overnode[sorted_index[0]]->a[i];
		group1->b[i] = overnode[sorted_index[0]]->b[i];
		group2->a[i] = overnode[sorted_index[M]]->a[i];
		group2->b[i] = overnode[sorted_index[M]]->b[i];
	}


	cut = m + k; 
	for (j=0; j<M+1; j++) {
  
		if (j < cut) {
  
			cal_MBR_node_node(group1->a, group1->b, group1, overnode[sorted_index[j]]);
                        
		}
		else {
                        	
			cal_MBR_node_node(group2->a, group2->b, group2, overnode[sorted_index[j]]);
  
		}

	}

	new_overlap_value = cal_overlap(group1, group2);
 
        
	if (k == 0) {   

		split_index = k;
		min_overlap_value = new_overlap_value;
		vol_at_index = cal_vol(group1->a, group1->b) + cal_vol(group2->a, group2->b);

  	}
	else {   
                
		if (new_overlap_value < min_overlap_value) {

			split_index = k;
			min_overlap_value = new_overlap_value;
			vol_at_index = cal_vol(group1->a, group1->b) + cal_vol(group2->a, group2->b);

		}
		else {

			new_vol = cal_vol(group1->a, group1->b) + cal_vol(group2->a, group2->b);
			if (new_vol < vol_at_index) {
				split_index = k;
			}

		}
  	}

  }

	
  for (i=0; i<dim; i++) {
	group1_chosen->a[i] = overnode[sorted_index[0]]->a[i];
	group1_chosen->b[i] = overnode[sorted_index[0]]->b[i];
	group2_chosen->a[i] = overnode[sorted_index[M]]->a[i];
	group2_chosen->b[i] = overnode[sorted_index[M]]->b[i];
  }

  //set_copy(group1_chosen->classGroup, overnode[sorted_index[0]]->classGroup);
  //set_copy(group2_chosen->classGroup, overnode[sorted_index[M]]->classGroup);

  classSet_copy(group1_chosen->classGroup, overnode[sorted_index[0]]->classGroup);
  classSet_copy(group2_chosen->classGroup, overnode[sorted_index[M]]->classGroup);



  cut = m + split_index;

  for (j=0; j<M+1; j++) {


	if (j < cut) {
                
		group1_chosen->ptr[j] = overnode[sorted_index[j]];

		overnode[sorted_index[j]]->parent = group1_chosen;
		cal_MBR_node_node(group1_chosen->a, group1_chosen->b, group1_chosen, overnode[sorted_index[j]]);

		//set_unionNonRedundant(group1_chosen->classGroup, overnode[sorted_index[j]]->classGroup);
		classSet_unionNonRedundant(group1_chosen->classGroup, overnode[sorted_index[j]]->classGroup);
	}


	else {

		group2_chosen->ptr[j-cut] = overnode[sorted_index[j]];
                overnode[sorted_index[j]]->parent = group2_chosen;
                cal_MBR_node_node(group2_chosen->a, group2_chosen->b, group2_chosen, overnode[sorted_index[j]]);

		//set_unionNonRedundant(group2_chosen->classGroup, overnode[sorted_index[j]]->classGroup);
		classSet_unionNonRedundant(group2_chosen->classGroup, overnode[sorted_index[j]]->classGroup);
	}

  }       


  group1_chosen->vacancy = M - cut;
  group2_chosen->vacancy = cut - 1;	// M - (M+1 - cut);

  tree_node_deallocate(group1);
  tree_node_deallocate(group2);
  free(sorted_index);

                       
  return;
        
} /* ChooseSplitIndex() */




void split(node_type *splitting_node, node_type *extra_node, node_type *node1, node_type *node2) 
{
  node_type **overnode;
  int axis_chosen;

  int i;

  overnode = (node_type **)malloc((M+1) * sizeof(node_type *));
  for (i=0; i<M; i++) {
        overnode[i] = splitting_node->ptr[i];
  }

  overnode[M] = extra_node;

  for(i=0; i < M; i++) {
        node1->ptr[i]=NULL;
        node2->ptr[i]=NULL;
  }

  axis_chosen = ChooseSplitAxis(overnode);

  ChooseSplitIndex(overnode, axis_chosen, node1, node2);

  node1->attribute = NODE;
  node1->parent = splitting_node->parent;
  node1->id = NO_ID;

  node2->attribute = NODE;
  node2->parent = splitting_node->parent;
  node2->id = NO_ID;

  free(overnode);

  return;

} /* split() */


//*** Ray has added the void return type of the function
void adjust_tree(node_type *splitting_node, int over_level, int old_level, node_type 
*node1, node_type *node2, node_type *root) 
{
  int split_child_no = 0;
  node_type *split_parent;

  int i, j;

  if(splitting_node->attribute == ROOT) { 

	/* The splitting node is the root */

	node1->parent = root;
	node2->parent = root;
	root->ptr[0] = node1;
	root->ptr[1] = node2;
	for (j=2; j<M; j++) 
		root->ptr[j] = NULL;
	root->parent = NULL;
	root->vacancy=M-2;
	root->attribute=ROOT;

	cal_MBR_node_node(root->a, root->b, node1, node2);

	//set_copy(root->classGroup, node1->classGroup);
	//set_unionNonRedundant(root->classGroup, node2->classGroup);

	classSet_copy(root->classGroup, node1->classGroup);
	classSet_unionNonRedundant(root->classGroup, node2->classGroup);


	extra_level++;
  }
  else { 

	/* The splitted is an intermediate node */

	split_parent = splitting_node->parent;

        for(i=0; i < M; i++) {
		if(split_parent->ptr[i] == splitting_node) { 
			split_child_no = i;
	             	break;
		}
        }

	tree_node_deallocate(splitting_node);	

	/* insert first node */
	split_parent->ptr[split_child_no] = node1;
	node1->parent = split_parent;

	adjust_MBR_delete(node1);

	/* insert second node */
	
	if(split_parent->vacancy != 0) { 

		/* no need to split again */
           	split_parent->ptr[M - split_parent->vacancy] = node2;

		node2->parent = split_parent;
		split_parent->vacancy--;

//		cal_MBR_node_node(split_parent->a, split_parent->b, node1, split_parent);
		cal_MBR_node_node(split_parent->a, split_parent->b, node2, split_parent);

		adjust_MBR(split_parent, node2->classGroup);

	}
       	else { 

		/* need to split again */
		overflow(split_parent, over_level-1, old_level, node2, root);

	}

  }



  return;

} /* adjust_tree */



void choose_leaf_level(node_type **node_found, node_type *current_node, int 
current_level, node_type *inserted_node, int desired_level) 
{
 
  int child_chosen;

  if (current_level == desired_level) {
        *node_found = current_node;
//printf("7\n");

        return;
  }
  
//printf("current_level %d desired_level %d\n", current_level, desired_level);

//printf("current_node->ptr[0]->ptr[0] %d\n", current_node->ptr[0]->ptr[0]);
  if (current_node->ptr[0]->ptr[0]->attribute != LEAF) {
  
//printf("8\n");
        child_chosen = least_area_enlarge(current_node, inserted_node);
   
  }
  else {
        
//printf("9\n");
        child_chosen = least_overlap_enlarge(current_node, inserted_node);
  
  }

//printf("5\n");

  current_node = current_node->ptr[child_chosen];
//printf("current_level %d desired_level %d\n", current_level, desired_level);
  choose_leaf_level(node_found, current_node, current_level+1, inserted_node, desired_level);

  
  return;
                        
} /* choose_leaf_level */




void reinsert(node_type *over_node, int over_level, node_type *extra_node, 
node_type *root) 
{
  node_type *node_found;
  node_type **overnode;
  double *value; 
  int *sorted_index;

  int i, start, stop;

  //**** Ray has fixed the following bugs
  //**** from "sizeof(double) * M+1" to "sizeof(double) * (M+1)"
  //**** from "sizeof(int) * M+1" to "sizeof(int) * (M+1)"
  value = (double *)malloc(sizeof(double) * (M+1));
  sorted_index = (int *)malloc(sizeof(int) * (M+1));

  overnode = (node_type **)malloc((M+1) * sizeof(node_type *));
  for (i=0; i<M; i++) {
        overnode[i] = over_node->ptr[i];
  }

  overnode[M] = extra_node;
  overnode[M]->parent = over_node;

  for (i=0; i<M+1; i++) {
	value[i] = Dist2(overnode[i], over_node);
	sorted_index[i] = i;
  }

  rtree_quicksort(sorted_index, value, 0, M);

  for (i=0; i<dim; i++) {
	over_node->a[i] = overnode[sorted_index[0]]->a[i];
	over_node->b[i] = overnode[sorted_index[0]]->b[i];
  }
  over_node->ptr[0] = overnode[sorted_index[0]];

//printf("1\n");

  stop = M+1-reinsert_p;

  for (i=1; i<stop; i++) {
	over_node->ptr[i] = overnode[sorted_index[i]];
	cal_MBR_node_node(over_node->a, over_node->b, over_node, overnode[sorted_index[i]]);
  }
  for (i=stop; i<M; i++)
	over_node->ptr[i] = NULL;

  over_node->vacancy = reinsert_p - 1;

  adjust_MBR_delete(over_node);

  start = M + 1 - reinsert_p;

  for (i=start; i<M+1; i++) {

	node_found = root;
	
	choose_leaf_level(&node_found, root, 0, overnode[sorted_index[i]], over_level+extra_level);
//printf("2\n");
 
	/* Test whether the node has room or not */
	if(node_found->vacancy!=0) {
 
		/* have room to insert the entry */
//printf("3\n");

		overnode[sorted_index[i]]->parent = node_found;

	        node_found->ptr[M - node_found->vacancy] = overnode[sorted_index[i]];

        	node_found->vacancy--;

	        adjust_MBR(overnode[sorted_index[i]], overnode[sorted_index[i]]->classGroup);
//printf("4\n");
	}
	else {

		overflow(node_found, over_level, over_level, overnode[sorted_index[i]], root);

	}

  }        


  return;
  
} /* reinsert */




void overflow(node_type *over_node, int over_level, int old_level, node_type
*extra_node, node_type *root)
{
  node_type *node1, *node2;

  if (over_level < old_level && over_level != 0) {
	reinsert(over_node, over_level, extra_node, root);

  }
  else {

	tree_node_allocate(&node1);
	tree_node_allocate(&node2);

        split(over_node, extra_node, node1, node2);

        adjust_tree(over_node, over_level, old_level, node1, node2, root);

//printf("overflow, go out adjust, extra_node->id %d\n", extra_node->id);


  }
        
  
  return;
  
}



//void insert_node(node_type *root, float *data, int seq_id, int dimNo, int classID)
void insert_node(node_type *root, float *data, int seq_id, myClassSet *projDimClass)
{
  node_type *node_found, *new_node, *data_node;
  int level_found;

  int i;

  //mySet *setClassID;
  myClassSet *setClassID;

  /******/
  /* I1 */
  /******/


  node_found = root;
  extra_level=0;

  tree_node_allocate(&data_node);
  for (i=0; i<dim; i++) {
	data_node->a[i] = data[i];
	data_node->b[i] = data[i];
  }
  //set_insert(data_node->classGroup, classID);
  //classSet_insert(data_node->classGroup, dimNo, classID);
  classSet_unionNonRedundant(data_node->classGroup, projDimClass);
  level_found = choose_leaf(&node_found, root, 0, data_node);
  /* Now, node_found is the pointer to the leaf chosen */


  /******/
  /* I2 */
  /******/
  
  /********************************************************/
  /* Add record to leaaf node:                            */
  /* If L has room for another entry, install the entry   */
  /* Otherwise invoke split() to split the node */
  /********************************************************/  
  
  /* Make a leaf node */

  tree_node_allocate(&new_node);


  for(i=0; i < dim; i++) { 
       	new_node->a[i]=data[i];
       	new_node->b[i]=data[i];
  }
  new_node->id = seq_id;		//data[dim] is the seq_id   
  new_node->attribute=LEAF;
  new_node->vacancy=M;
  new_node->parent=node_found;
  //set_insert(new_node->classGroup, classID);
  //classSet_insert(new_node->classGroup, dimNo, classID);
  classSet_unionNonRedundant(new_node->classGroup, projDimClass);
  for(i=0; i < M; i++)
     new_node->ptr[i]=NULL;

  /* Test whether the node has room or not */
  if(node_found->vacancy!=0) { 

//printf("insert new->id %d\n", seq_id);

	/* have room to insert the entry */

	node_found->ptr[M - node_found->vacancy] = new_node;

      	node_found->vacancy--;
    
	//setClassID = set_new();
	//set_insert(setClassID, classID);
	setClassID = classSet_new();
	//classSet_insert(setClassID, dimNo, classID);
	classSet_unionNonRedundant(setClassID, projDimClass);

	adjust_MBR(new_node, setClassID);

	//set_free(setClassID);
	//free(setClassID);
	classSet_free(setClassID);
	free(setClassID);
  }
  else {

	overflow(node_found, level_found, level_found+1, new_node, root);

  }


  return;

} /* insert_node */
  




/********************/
/* build_tree():    */
/* build the R-tree */
/********************/
      
void build_tree(node_type **root, float **data, int no_data)
{ 
  int i, j;


  /* make tree root node first */

  tree_node_allocate(root);

  for(i=0; i < dim; i++) { 
	(*root)->b[i] = (float)(-1 * INT_MAX);
       	(*root)->a[i] = (float)(INT_MAX);
  }
  (*root)->id = -3;		//NO_ID
  (*root)->attribute=ROOT;
  (*root)->vacancy=M;   
  (*root)->parent=NULL; 
  for(j=0; j < M; j++)
     	(*root)->ptr[j]=NULL;


  /* add data to the tree */
  for(i=0; i<no_data; i++) { 

#ifdef DEBUG
if (i+1%1000==0)
	printf("insert data %d\n", i+1);
#endif

      	//insert_node(*root, data[i], i, classID);  // i is the seq id.


	free(data[i]);
  }

  free(data);

} /*build_tree */


int make_data(char *positionfile, float ***data)
{ 
  int no_data;
  int i,j;
  FILE *fp_position;

  fp_position=fopen(positionfile,"r");  

  no_data = no_histogram;

  (*data) = (float **)malloc(sizeof(float*) * no_data);
  for(i=0; i<no_data; i++)
        (*data)[i] = (float *)malloc(sizeof(float) * dim);

  for(i=0; i<no_data; i++)
	for (j=0; j<dim; j++)
		fscanf(fp_position,"%f",&((*data)[i][j]));

  fclose(fp_position);

  return(no_data);

} /* make_data */



void write_leaf_node(node_type *node, FILE *fp)
{
  int i;
                
  for (i = 0; i<dim; i++)
        fprintf(fp, "%f\n", (node->a)[i]);
 
  for (i = 0; i<dim; i++)
        fprintf(fp, "%f\n", (node->b)[i]);
 
  fprintf(fp, "%d\n", node->attribute);
  fprintf(fp, "%d\n", node->id);
  fprintf(fp, "%d\n", node->vacancy);

  return;

}


void write_inter_node(node_type *node, FILE *fp)
{
  int i, count;

  for (i = 0; i<dim; i++) 
	fprintf(fp, "%f\n", (node->a)[i]);

  for (i = 0; i<dim; i++)  
        fprintf(fp, "%f\n", (node->b)[i]);

  fprintf(fp, "%d\n", node->attribute);
  fprintf(fp, "%d\n", node->vacancy);

  count = M - node->vacancy;
  for (i=0; i<count; i++) {
	if (node->ptr[i]->attribute != LEAF)
		write_inter_node(node->ptr[i], fp);
	else
		write_leaf_node(node->ptr[i], fp);
  }

  return;

}
  

void save_rtree(node_type *root)
{
  FILE *fp;

  fp = fopen(SAVE_RTREE_FILE, "w");
  
  write_inter_node(root, fp);

  fclose(fp);
 
  return;

}

rtreeGroup *rtreeGroup_new()
{
	rtreeGroup *returnValue;

	returnValue = (rtreeGroup *) malloc(sizeof(rtreeGroup));
	memset(returnValue, 0, sizeof(rtreeGroup));

	return returnValue;
}

void discard_inter_node(node_type **aNode)
{
	int count;
	int i;

	node_type *currNode;

	currNode = (*aNode);
	if (currNode->attribute != LEAF)
	{

		// free children
		count = M - currNode->vacancy;
		for (i = 0; i < count; i++)
		{
			discard_inter_node(&(currNode->ptr[i]));
		}
		

		// free itself
		free(currNode->ptr);
		free(currNode->a);
		free(currNode->b);
		free(currNode);
	}
}

void discard_rtree(node_type **root)
{
	discard_inter_node(root);
}

void discard_rtreeGroup(rtreeGroup *anRtreeGroup)
{
	node_type **rtreeList;

	int i;

	rtreeList = anRtreeGroup->rtreeList;

	for (i = 0; i < anRtreeGroup->noOfRtree; i++)
	{
		discard_rtree(&(rtreeList[i]));
	}
	free(rtreeList);
}

void count_rtreeSize_node(node_type **aNode, int *noOfNode)
{
	int count; 
	int i;

	node_type *currNode;

	currNode = (*aNode);
	if (currNode->attribute != LEAF)
	{
		count = M - currNode->vacancy;
		for (i = 0; i < count; i++)
		{
			count_rtreeSize_node(&(currNode->ptr[i]), noOfNode);
		}
	}

	(*noOfNode)++;
}

int count_rtreeSize(node_type **root)
{
	int returnValue;

	returnValue = 0;

	count_rtreeSize_node(root, &returnValue);

	return returnValue;
}

void printUsage(char *name)
{
	printf("%s <no. of points> <dimensionality> <data file> <rstree file> <m> <M> <reinsert-p>\n", name);
}

int main2(int argc, char **argv)
{ 
  int no_data;
  float **data;
  configR_type configR;
  node_type *root;

  //initialize(&configR);   
  if (argc != 8)
  {
	  printUsage(argv[0]);
	  exit(0);
  }
  initialize_argument(&configR, argv);

  no_data=make_data(configR.positionfile, &data);

  build_tree(&root, data, no_data);
  save_rtree(root);

  printf("Discarding rtree...\n");
  discard_rtree(&root);

  return(0);  
} /* main */  

