#include "list.h"

list* list_new()
{
	list *l;
	l = (list *)malloc(sizeof(list));
	l->size = 0;
	l->start = NULL;
	l->end = NULL;
	return l;
}

list_entry* list_entry_new(int val)
{
	list_entry *le;
	le = (list_entry *)malloc(sizeof(list_entry));
	le->elt = val;
	le->next = NULL;
	le->prev = NULL;
	return le;
}
void list_addAtHead(list *l, int val)
{

	list_entry *le = list_entry_new(val);
	if(l->size==0)
	{
		l->start = le;
		l->end= le;
		l->size++;
	}
	else
	{
		if(l->start->elt>val)
		{
			l->start->prev = le;
			le->next = l->start;
			l->start = le;
			l->size++;
		}
	}
}
void list_addAtTail(list *l, int val)
{

	list_entry *le = list_entry_new(val);
	if(l->size==0)
	{
		l->start = le;
		l->end= le;
		l->size++;
	}
	else
	{
		if(l->end->elt<val)
		{
			l->end->next = le;
			le->prev = l->end;
			l->end = le;
			l->size++;
		}
	}
}

void list_add(list *l, int val)
{
	list_entry *le = list_entry_new(val);
	if(l->size==0)
	{
		l->start = le;
		l->end= le;
		l->size++;
	}
	else
	{
		list_entry *currE = l->end;
		if(currE->elt>val)
		{
			while(currE->prev && (currE->prev->elt>val))
			{
				currE = currE->prev;
			}
			if(currE==l->start)
			{
				currE->prev= le;
				le->next = currE;
				l->start= le;
				l->size++;
			}
			else if(currE->prev->elt<val)
			{
				le->prev = currE->prev;
				currE->prev->next = le;
				le->next = currE;
				currE->prev = le;
				l->size++;
			}
		}
		else if(currE->elt<val)
		{
			l->end->next = le;
			le->prev = l->end;
			l->end = le;
			l->size++;
		}
	}
}
		
void list_free(list *l)
{
	list_entry *currE;
	currE = l->start;
	while(currE)
	{
		list_entry *nextE;
		nextE = currE->next;
		free(currE);
		currE = nextE;
	}
	free(l);
}

void list_print(list *l, FILE *stream)
{
	list_entry *currE;
	currE = l->start;
	while(currE)
	{
		fprintf(stream, "%d ", currE->elt);
		currE = currE->next;
	}
	putchar('\n');
}

list* list_union(list *l1, list *l2)
{
	return list_merage(1, l1, l2);
}

list* list_intersect(list *l1, list *l2)
{
	return list_merage(0, l1, l2);
}

list* list_merage(int mode, list *l1, list *l2)
{
	list *l = list_new();
	int a, b;
	list_entry *aE, *bE;
	a = 0;
	b = 0;
	aE = l1->start;
	bE = l2->start;
	while(a<l1->size && b<l2->size)
	{
		if(aE->elt == bE->elt)
		{
			list_addAtTail(l, aE->elt);
			a++;b++;
			aE = aE->next;
			bE = bE->next;
		}
		else if(aE->elt < bE->elt)
		{
			if(mode)
			{
				list_addAtTail(l, aE->elt);
			}
			a++;
			aE = aE->next;
		}
		else
		{
			if(mode)
			{
				list_addAtTail(l, bE->elt);
			}
			b++;
			bE = bE->next;
		}
	}
	if(mode)
	{
		while(a<l1->size)
		{
			list_addAtTail(l, aE->elt);
			a++;
			aE = aE->next;
		}
		while(b<l2->size)
		{
			list_addAtTail(l, bE->elt);
			b++;
			bE = bE->next;
		}
	}

	return l;
}

list* list_minus(list *l1, list *l2)
{
	list *l = list_new();
	int a, b;
	list_entry *aE, *bE;
	a = 0;
	b = 0;
	aE = l1->start;
	bE = l2->start;
	while(a<l1->size && b<l2->size)
	{
		if(aE->elt == bE->elt)
		{
			a++;b++;
			aE = aE->next;
			bE = bE->next;
		}
		else if(aE->elt < bE->elt)
		{
			list_addAtTail(l, aE->elt);
			a++;
			aE = aE->next;
		}
		else
		{
			b++;
			bE = bE->next;
		}
	}
	while(a<l1->size)
	{
		list_addAtTail(l, aE->elt);
		a++;
		aE = aE->next;
	}
	return l;
}
